unicode-i18n.h for extraction by gettext
THIS IS A GENERATED FILE. CHANGES WILL BE OVERWRITTEN.
Generated by ./gen-guch-unicode-tables.pl
Generated from UCD version 17.0.0

N_("Adlam")
N_("Aegean Numbers")
N_("Ahom")
N_("Alchemical Symbols")
N_("Alphabetic Presentation Forms")
N_("Anatolian Hieroglyphs")
N_("Ancient Greek Musical Notation")
N_("Ancient Greek Numbers")
N_("Ancient Symbols")
N_("Arabic")
N_("Arabic Extended-A")
N_("Arabic Extended-B")
N_("Arabic Extended-C")
N_("Arabic Mathematical Alphabetic Symbols")
N_("Arabic Presentation Forms-A")
N_("Arabic Presentation Forms-B")
N_("Arabic Supplement")
N_("Armenian")
N_("Arrows")
N_("Avestan")
N_("Balinese")
N_("Bamum")
N_("Bamum Supplement")
N_("Basic Latin")
N_("Bassa Vah")
N_("Batak")
N_("Bengali")
N_("Beria Erfe")
N_("Bhaiksuki")
N_("Block Elements")
N_("Bopomofo")
N_("Bopomofo Extended")
N_("Box Drawing")
N_("Brahmi")
N_("Braille Patterns")
N_("Buginese")
N_("Buhid")
N_("Byzantine Musical Symbols")
N_("CJK Compatibility")
N_("CJK Compatibility Forms")
N_("CJK Compatibility Ideographs")
N_("CJK Compatibility Ideographs Supplement")
N_("CJK Radicals Supplement")
N_("CJK Strokes")
N_("CJK Symbols and Punctuation")
N_("CJK Unified Ideographs")
N_("CJK Unified Ideographs Extension A")
N_("CJK Unified Ideographs Extension B")
N_("CJK Unified Ideographs Extension C")
N_("CJK Unified Ideographs Extension D")
N_("CJK Unified Ideographs Extension E")
N_("CJK Unified Ideographs Extension F")
N_("CJK Unified Ideographs Extension G")
N_("CJK Unified Ideographs Extension H")
N_("CJK Unified Ideographs Extension I")
N_("CJK Unified Ideographs Extension J")
N_("Carian")
N_("Caucasian Albanian")
N_("Chakma")
N_("Cham")
N_("Cherokee")
N_("Cherokee Supplement")
N_("Chess Symbols")
N_("Chorasmian")
N_("Combining Diacritical Marks")
N_("Combining Diacritical Marks Extended")
N_("Combining Diacritical Marks Supplement")
N_("Combining Diacritical Marks for Symbols")
N_("Combining Half Marks")
N_("Common Indic Number Forms")
N_("Control Pictures")
N_("Coptic")
N_("Coptic Epact Numbers")
N_("Counting Rod Numerals")
N_("Cuneiform")
N_("Cuneiform Numbers and Punctuation")
N_("Currency Symbols")
N_("Cypriot Syllabary")
N_("Cypro-Minoan")
N_("Cyrillic")
N_("Cyrillic Extended-A")
N_("Cyrillic Extended-B")
N_("Cyrillic Extended-C")
N_("Cyrillic Extended-D")
N_("Cyrillic Supplement")
N_("Deseret")
N_("Devanagari")
N_("Devanagari Extended")
N_("Devanagari Extended-A")
N_("Dingbats")
N_("Dives Akuru")
N_("Dogra")
N_("Domino Tiles")
N_("Duployan")
N_("Early Dynastic Cuneiform")
N_("Egyptian Hieroglyph Format Controls")
N_("Egyptian Hieroglyphs")
N_("Egyptian Hieroglyphs Extended-A")
N_("Elbasan")
N_("Elymaic")
N_("Emoticons")
N_("Enclosed Alphanumeric Supplement")
N_("Enclosed Alphanumerics")
N_("Enclosed CJK Letters and Months")
N_("Enclosed Ideographic Supplement")
N_("Ethiopic")
N_("Ethiopic Extended")
N_("Ethiopic Extended-A")
N_("Ethiopic Extended-B")
N_("Ethiopic Supplement")
N_("Garay")
N_("General Punctuation")
N_("Geometric Shapes")
N_("Geometric Shapes Extended")
N_("Georgian")
N_("Georgian Extended")
N_("Georgian Supplement")
N_("Glagolitic")
N_("Glagolitic Supplement")
N_("Gothic")
N_("Grantha")
N_("Greek Extended")
N_("Greek and Coptic")
N_("Gujarati")
N_("Gunjala Gondi")
N_("Gurmukhi")
N_("Gurung Khema")
N_("Halfwidth and Fullwidth Forms")
N_("Hangul Compatibility Jamo")
N_("Hangul Jamo")
N_("Hangul Jamo Extended-A")
N_("Hangul Jamo Extended-B")
N_("Hangul Syllables")
N_("Hanifi Rohingya")
N_("Hanunoo")
N_("Hatran")
N_("Hebrew")
N_("High Private Use Surrogates")
N_("High Surrogates")
N_("Hiragana")
N_("IPA Extensions")
N_("Ideographic Description Characters")
N_("Ideographic Symbols and Punctuation")
N_("Imperial Aramaic")
N_("Indic Siyaq Numbers")
N_("Inscriptional Pahlavi")
N_("Inscriptional Parthian")
N_("Javanese")
N_("Kaithi")
N_("Kaktovik Numerals")
N_("Kana Extended-A")
N_("Kana Extended-B")
N_("Kana Supplement")
N_("Kanbun")
N_("Kangxi Radicals")
N_("Kannada")
N_("Katakana")
N_("Katakana Phonetic Extensions")
N_("Kawi")
N_("Kayah Li")
N_("Kharoshthi")
N_("Khitan Small Script")
N_("Khmer")
N_("Khmer Symbols")
N_("Khojki")
N_("Khudawadi")
N_("Kirat Rai")
N_("Lao")
N_("Latin Extended Additional")
N_("Latin Extended-A")
N_("Latin Extended-B")
N_("Latin Extended-C")
N_("Latin Extended-D")
N_("Latin Extended-E")
N_("Latin Extended-F")
N_("Latin Extended-G")
N_("Latin-1 Supplement")
N_("Lepcha")
N_("Letterlike Symbols")
N_("Limbu")
N_("Linear A")
N_("Linear B Ideograms")
N_("Linear B Syllabary")
N_("Lisu")
N_("Lisu Supplement")
N_("Low Surrogates")
N_("Lycian")
N_("Lydian")
N_("Mahajani")
N_("Mahjong Tiles")
N_("Makasar")
N_("Malayalam")
N_("Mandaic")
N_("Manichaean")
N_("Marchen")
N_("Masaram Gondi")
N_("Mathematical Alphanumeric Symbols")
N_("Mathematical Operators")
N_("Mayan Numerals")
N_("Medefaidrin")
N_("Meetei Mayek")
N_("Meetei Mayek Extensions")
N_("Mende Kikakui")
N_("Meroitic Cursive")
N_("Meroitic Hieroglyphs")
N_("Miao")
N_("Miscellaneous Mathematical Symbols-A")
N_("Miscellaneous Mathematical Symbols-B")
N_("Miscellaneous Symbols")
N_("Miscellaneous Symbols Supplement")
N_("Miscellaneous Symbols and Arrows")
N_("Miscellaneous Symbols and Pictographs")
N_("Miscellaneous Technical")
N_("Modi")
N_("Modifier Tone Letters")
N_("Mongolian")
N_("Mongolian Supplement")
N_("Mro")
N_("Multani")
N_("Musical Symbols")
N_("Myanmar")
N_("Myanmar Extended-A")
N_("Myanmar Extended-B")
N_("Myanmar Extended-C")
N_("N'Ko")
N_("Nabataean")
N_("Nag Mundari")
N_("Nandinagari")
N_("New Tai Lue")
N_("Newa")
N_("Number Forms")
N_("Nushu")
N_("Nyiakeng Puachue Hmong")
N_("Ogham")
N_("Ol Chiki")
N_("Ol Onal")
N_("Old Hungarian")
N_("Old Italic")
N_("Old North Arabian")
N_("Old Permic")
N_("Old Persian")
N_("Old Sogdian")
N_("Old South Arabian")
N_("Old Turkic")
N_("Old Uyghur")
N_("Optical Character Recognition")
N_("Oriya")
N_("Ornamental Dingbats")
N_("Osage")
N_("Osmanya")
N_("Ottoman Siyaq Numbers")
N_("Pahawh Hmong")
N_("Palmyrene")
N_("Pau Cin Hau")
N_("Phags-pa")
N_("Phaistos Disc")
N_("Phoenician")
N_("Phonetic Extensions")
N_("Phonetic Extensions Supplement")
N_("Playing Cards")
N_("Private Use Area")
N_("Psalter Pahlavi")
N_("Rejang")
N_("Rumi Numeral Symbols")
N_("Runic")
N_("Samaritan")
N_("Saurashtra")
N_("Sharada")
N_("Sharada Supplement")
N_("Shavian")
N_("Shorthand Format Controls")
N_("Siddham")
N_("Sidetic")
N_("Sinhala")
N_("Sinhala Archaic Numbers")
N_("Small Form Variants")
N_("Small Kana Extension")
N_("Sogdian")
N_("Sora Sompeng")
N_("Soyombo")
N_("Spacing Modifier Letters")
N_("Specials")
N_("Sundanese")
N_("Sundanese Supplement")
N_("Sunuwar")
N_("Superscripts and Subscripts")
N_("Supplemental Arrows-A")
N_("Supplemental Arrows-B")
N_("Supplemental Arrows-C")
N_("Supplemental Mathematical Operators")
N_("Supplemental Punctuation")
N_("Supplemental Symbols and Pictographs")
N_("Supplementary Private Use Area-A")
N_("Supplementary Private Use Area-B")
N_("Sutton SignWriting")
N_("Syloti Nagri")
N_("Symbols and Pictographs Extended-A")
N_("Symbols for Legacy Computing")
N_("Symbols for Legacy Computing Supplement")
N_("Syriac")
N_("Syriac Supplement")
N_("Tagalog")
N_("Tagbanwa")
N_("Tags")
N_("Tai Le")
N_("Tai Tham")
N_("Tai Viet")
N_("Tai Xuan Jing Symbols")
N_("Tai Yo")
N_("Takri")
N_("Tamil")
N_("Tamil Supplement")
N_("Tangsa")
N_("Tangut")
N_("Tangut Components")
N_("Tangut Components Supplement")
N_("Tangut Supplement")
N_("Telugu")
N_("Thaana")
N_("Thai")
N_("Tibetan")
N_("Tifinagh")
N_("Tirhuta")
N_("Todhri")
N_("Tolong Siki")
N_("Toto")
N_("Transport and Map Symbols")
N_("Tulu-Tigalari")
N_("Ugaritic")
N_("Unified Canadian Aboriginal Syllabics")
N_("Unified Canadian Aboriginal Syllabics Extended")
N_("Unified Canadian Aboriginal Syllabics Extended-A")
N_("Vai")
N_("Variation Selectors")
N_("Variation Selectors Supplement")
N_("Vedic Extensions")
N_("Vertical Forms")
N_("Vithkuqi")
N_("Wancho")
N_("Warang Citi")
N_("Yezidi")
N_("Yi Radicals")
N_("Yi Syllables")
N_("Yijing Hexagram Symbols")
N_("Zanabazar Square")
N_("Znamenny Musical Notation")

N_("Adlam")
N_("Ahom")
N_("Anatolian Hieroglyphs")
N_("Arabic")
N_("Armenian")
N_("Avestan")
N_("Balinese")
N_("Bamum")
N_("Bassa Vah")
N_("Batak")
N_("Bengali")
N_("Beria Erfe")
N_("Bhaiksuki")
N_("Bopomofo")
N_("Brahmi")
N_("Braille")
N_("Buginese")
N_("Buhid")
N_("Canadian Aboriginal")
N_("Carian")
N_("Caucasian Albanian")
N_("Chakma")
N_("Cham")
N_("Cherokee")
N_("Chorasmian")
N_("Common")
N_("Coptic")
N_("Cuneiform")
N_("Cypriot")
N_("Cypro Minoan")
N_("Cyrillic")
N_("Deseret")
N_("Devanagari")
N_("Dives Akuru")
N_("Dogra")
N_("Duployan")
N_("Egyptian Hieroglyphs")
N_("Elbasan")
N_("Elymaic")
N_("Ethiopic")
N_("Garay")
N_("Georgian")
N_("Glagolitic")
N_("Gothic")
N_("Grantha")
N_("Greek")
N_("Gujarati")
N_("Gunjala Gondi")
N_("Gurmukhi")
N_("Gurung Khema")
N_("Han")
N_("Hangul")
N_("Hanifi Rohingya")
N_("Hanunoo")
N_("Hatran")
N_("Hebrew")
N_("Hiragana")
N_("Imperial Aramaic")
N_("Inherited")
N_("Inscriptional Pahlavi")
N_("Inscriptional Parthian")
N_("Javanese")
N_("Kaithi")
N_("Kannada")
N_("Katakana")
N_("Kawi")
N_("Kayah Li")
N_("Kharoshthi")
N_("Khitan Small Script")
N_("Khmer")
N_("Khojki")
N_("Khudawadi")
N_("Kirat Rai")
N_("Lao")
N_("Latin")
N_("Lepcha")
N_("Limbu")
N_("Linear A")
N_("Linear B")
N_("Lisu")
N_("Lycian")
N_("Lydian")
N_("Mahajani")
N_("Makasar")
N_("Malayalam")
N_("Mandaic")
N_("Manichaean")
N_("Marchen")
N_("Masaram Gondi")
N_("Medefaidrin")
N_("Meetei Mayek")
N_("Mende Kikakui")
N_("Meroitic Cursive")
N_("Meroitic Hieroglyphs")
N_("Miao")
N_("Modi")
N_("Mongolian")
N_("Mro")
N_("Multani")
N_("Myanmar")
N_("N'Ko")
N_("Nabataean")
N_("Nag Mundari")
N_("Nandinagari")
N_("New Tai Lue")
N_("Newa")
N_("Nushu")
N_("Nyiakeng Puachue Hmong")
N_("Ogham")
N_("Ol Chiki")
N_("Ol Onal")
N_("Old Hungarian")
N_("Old Italic")
N_("Old North Arabian")
N_("Old Permic")
N_("Old Persian")
N_("Old Sogdian")
N_("Old South Arabian")
N_("Old Turkic")
N_("Old Uyghur")
N_("Oriya")
N_("Osage")
N_("Osmanya")
N_("Pahawh Hmong")
N_("Palmyrene")
N_("Pau Cin Hau")
N_("Phags Pa")
N_("Phoenician")
N_("Psalter Pahlavi")
N_("Rejang")
N_("Runic")
N_("Samaritan")
N_("Saurashtra")
N_("Sharada")
N_("Shavian")
N_("Siddham")
N_("Sidetic")
N_("Signwriting")
N_("Sinhala")
N_("Sogdian")
N_("Sora Sompeng")
N_("Soyombo")
N_("Sundanese")
N_("Sunuwar")
N_("Syloti Nagri")
N_("Syriac")
N_("Tagalog")
N_("Tagbanwa")
N_("Tai Le")
N_("Tai Tham")
N_("Tai Viet")
N_("Tai Yo")
N_("Takri")
N_("Tamil")
N_("Tangsa")
N_("Tangut")
N_("Telugu")
N_("Thaana")
N_("Thai")
N_("Tibetan")
N_("Tifinagh")
N_("Tirhuta")
N_("Todhri")
N_("Tolong Siki")
N_("Toto")
N_("Tulu Tigalari")
N_("Ugaritic")
N_("Unknown")
N_("Vai")
N_("Vithkuqi")
N_("Wancho")
N_("Warang Citi")
N_("Yezidi")
N_("Yi")
N_("Zanabazar Square")
